<?php

declare(strict_types=1);

namespace Drupal\Tests\rdf_sync\Kernel;

use Drupal\Core\Serialization\Yaml;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\KernelTests\KernelTestBase;
use Drupal\node\Entity\Node;
use Drupal\node\NodeInterface;
use Drupal\rdf_sync\Model\SyncMode;
use Drupal\rdf_sync\RdfSyncConnection;
use Drupal\rdf_sync_test\Category;
use Drupal\taxonomy\TermInterface;
use Drupal\Tests\field\Traits\EntityReferenceFieldCreationTrait;
use PHPUnit\Framework\Assert;

/**
 * Tests RDF serialization.
 *
 * @group rdf_sync
 */
class RdfSyncServiceTest extends KernelTestBase {

  use EntityReferenceFieldCreationTrait;

  /**
   * Testing taxonomy term.
   */
  protected TermInterface $term;

  /**
   * Testing node.
   */
  protected NodeInterface $node;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'field',
    'filter',
    'node',
    'rdf_sync',
    'rdf_sync_test',
    'serialization',
    'system',
    'taxonomy',
    'text',
    'user',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('taxonomy_term');
    $this->installEntitySchema('node');
    $this->installEntitySchema('user');
    $this->installConfig(['rdf_sync_test', 'rdf_sync']);

    $this->container->get('config.factory')->getEditable('rdf_sync.settings')
      ->set('endpoint.host', 'virtuoso')
      ->save();

    // Add a body field.
    FieldStorageConfig::create([
      'entity_type' => 'node',
      'type' => 'text_with_summary',
      'field_name' => 'body',
      'cardinality' => 1,
    ])->save();
    FieldConfig::create([
      'entity_type' => 'node',
      'bundle' => 'page',
      'field_name' => 'body',
      'label' => $this->randomString(),
    ])
      // Using the computed 'processed' column instead of 'value' to get the
      // text format applied.
      ->setThirdPartySetting('rdf_sync', 'processed', [
        'predicate' => 'http://example.com/page/body',
        'type' => NULL,
      ])->save();

    $this->createEntityReferenceField('node', 'page', 'category', $this->randomString(), 'taxonomy_term', 'default', ['target_bundles' => ['category']]);
    FieldConfig::loadByName('node', 'page', 'category')
      ->setThirdPartySetting('rdf_sync', 'target_id', [
        'predicate' => 'http://example.com/page/category',
        'type' => 'resource',
      ])->save();

    // Start on clean.
    $this->container->get('entity_field.manager')->clearCachedFieldDefinitions();
    $this->container->get('entity_type.bundle.info')->clearCachedBundles();

    $this->createEntities();
  }

  /**
   * @covers \Drupal\rdf_sync\Normalizer\RdfSyncNormalizer
   * @covers \Drupal\rdf_sync\Encoder\RdfSyncEncoder
   *
   * @dataProvider providerTestSerialization
   */
  public function testSerialization(string $format, array $expectedSerialization): void {
    $actualTermSerialization = trim($this->container->get('serializer')->serialize($this->term, $format));
    $this->assertSame($expectedSerialization['taxonomy_term'], $actualTermSerialization);
    $actualNodeSerialization = trim($this->container->get('serializer')->serialize($this->node, $format));
    $this->assertSame($expectedSerialization['node'], $actualNodeSerialization);
  }

  /**
   * @covers \Drupal\rdf_sync\RdfSyncSynchronizer
   */
  public function testSynchronization(): void {
    /** @var \Drupal\rdf_sync\RdfSyncSynchronizer $synchronizer */
    $synchronizer = $this->container->get('rdf_sync.synchronizer');
    $synchronizer->synchronize([$this->term, $this->node], 'insert', SyncMode::INSTANT);
    $this->assertTriples('http://data', [
      '<http://example.com/category/id/1> <http://www.w3.org/2004/02/skos/core#inScheme> <http://example.com/category>',
      '<http://example.com/category/id/1> <http://www.w3.org/2004/02/skos/core#prefLabel> "Series"@en',
      '<http://example.com/category/id/1> <http://www.w3.org/2004/02/skos/core#definition> "7 seasons"@en',
      '<http://example.com/category/id/1> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://www.w3.org/2004/02/skos/core#Concept>',
      '<http://example.com/page/id/1> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://example.com/page>',
      '<http://example.com/page/id/1> <http://example.com/page/title> "Mad Men"@en',
      '<http://example.com/page/id/1> <http://example.com/page/category> <http://example.com/category/id/1>',
      '<http://example.com/page/id/1> <http://example.com/page/body> "Strong is stripped while <em>Emphasis</em> is kept"@en',
    ]);
  }

  /**
   * Provides test cases for self::testSerialization())
   *
   * @return array
   *   A list of testing cases.
   *
   * @see self::testSerialization()
   */
  public function providerTestSerialization(): array {
    $cases = Yaml::decode(file_get_contents(__DIR__ . '/../../fixtures/serialization.yml'));
    return array_map(function (string $format) use ($cases): array {
      return [$format, $cases[$format]];
    }, array_keys($cases));
  }

  /**
   * Creates testing entities.
   */
  protected function createEntities(): void {
    $this->term = Category::create([
      'vid' => 'category',
      'name' => 'Series',
      'description' => [
        'value' => '<b>7</b> seasons',
        'format' => 'html',
      ],
      'uri' => 'http://example.com/category/id/1',
    ]);
    $this->node = Node::create([
      'type' => 'page',
      'title' => 'Mad Men',
      'uri' => 'http://example.com/page/id/1',
      'body' => [
        'value' => '<strong>Strong</strong> is stripped while <em>Emphasis</em> is kept',
        'format' => 'html',
      ],
      'category' => $this->term,
    ]);
  }

  /**
   * Asserts that an RDF graph contans a list of triples.
   *
   * @param string $graph
   *   The graph URI.
   * @param string[] $triples
   *   A list of triples. Each entry is a ntriples serializezed triple.
   */
  protected function assertTriples(string $graph, array $triples) {
    $failures = [];
    foreach ($triples as $triple) {
      if (!RdfSyncConnection::query("ASK WHERE { GRAPH <$graph> { $triple } }")->getBoolean()) {
        $failures[] = $triple;
      }
    }
    Assert::assertEmpty($failures);
  }

}
